#include <msg/msg.h>
#include <xbt/dict.h>
#include <stdio.h>
#include <stdlib.h>
#include "grid_agents.h"
#include <string.h>
#include <math.h>

/* Create a log channel to have nice outputs. */
#include "xbt/log.h"
#include "xbt/asserts.h"

XBT_LOG_NEW_DEFAULT_CATEGORY(remi, "Messages specific for the main example");
XBT_LOG_NEW_CATEGORY(alvin, "Debug Messages for Arnaud");
XBT_LOG_NEW_CATEGORY(remi_bounds, "Debug on feasibility");

XBT_LOG_NEW_CATEGORY(all_vars, "Printing everything, for further use");
XBT_LOG_NEW_CATEGORY(rho_vars, "Printing only rhos");

// const char channel_buf[10000];
static __inline__ const char *CHANNEL(const char *channel_buf, const char* prefix, int app_id,
		m_host_t sender, m_host_t receiver) {
	snprintf((char * __restrict__) channel_buf, 1000, "%s|%d:%s->%s", prefix,
			app_id, sender->name, receiver->name);
	return channel_buf;
}

#define VOIRS(expr) XBT_CDEBUG(alvin,"  {" #expr " = %s }", expr)
#define VOIRP(expr) XBT_CDEBUG(alvin,"  {" #expr " = %p }", expr)
#define VOIRD(expr) XBT_CDEBUG(alvin,"  {" #expr " = %d }", expr)
#define VOIRF(expr) XBT_CDEBUG(alvin,"  {" #expr " = %f }", expr)
#define VOIRG(expr) XBT_CDEBUG(alvin,"  {" #expr " = %.8g }", expr)


static void trace_variable(double timestamp, const char *slave1, const char *slave2,
		const char *var_name, double value);

lambda_item_t *lambda_table = NULL;
int nb_lambda;
mu_item_t *mu_table = NULL;
int nb_mu;

int NUMBER_OF_APPLICATIONS = 1;

xbt_dict_t appli_dict = NULL;
xbt_dict_t host_dict = NULL;

FILE *bigfile = NULL;

FILE *readfile = NULL;
xbt_dict_t readfile_dict = NULL;

xbt_fifo_t hosts_waiting_for_bench = NULL;

void read_file_to_dict(int numero) {

	char variable[128];
	char string_time[128];
	char host_app[128];
	char value[128];
	int natural_time = 0;

	if (!readfile_dict)
		readfile_dict = xbt_dict_new();
	xbt_assert((bigfile = fopen("output.dat", "w")),
			"Unable to open output.dat");
	fclose(bigfile);
	if (numero < 0)
		return;

	readfile = fopen(parameters.input_file, "r");

	if (readfile != NULL) {

		xbt_assert((bigfile = fopen("output.dat", "w")),
				"Unable to open output.dat");
		while (fscanf(readfile, "%s %s %s %s", variable, string_time, host_app,
				value) != EOF) {

			xbt_assert(sscanf(string_time, "%i", &natural_time) == 1,
					"Not a time %s", string_time);

			if (natural_time == numero) {
				char chaine[128];
				sprintf(chaine, "%s %i %s", variable, natural_time, host_app);
				char *newchaine = strdup(chaine);
				char *newvalue = strdup(value);
				xbt_dict_set(readfile_dict, newchaine, newvalue, NULL);
			}

			if (natural_time < numero) {
				char chaine[128];
				sprintf(chaine, "%s %i %s %s\n", variable, natural_time,
						host_app, value);
				XBT_CVERB(all_vars, "%s", chaine);
				//fprintf(bigfile, chaine);
			}

		}
		fclose(bigfile);
		fclose(readfile);
	} else {
		printf("Unable to open : input.dat\n");
	}

}

static void register_application(int app) {
	char name[30];

	if (!appli_dict) {
		appli_dict = xbt_dict_new();
	}
	sprintf(name, "%d", app);
	xbt_dict_set(appli_dict, name, (void *) 1, NULL);
}

static void register_host(void) {
	if (!host_dict) {
		host_dict = xbt_dict_new();
	}
	xbt_dict_set(host_dict, MSG_host_self()->name, (void *) 1, NULL);
}

static void create_mu_and_lambda_table(void) {
	int noa = 0;
	int no_cpu = 0;
	int i;

	if (lambda_table)
		return;

	xbt_assert((bigfile = fopen("output.dat", "a")),
			"Unable to open output.dat");

	noa = xbt_dict_length(appli_dict);
	no_cpu = xbt_dict_length(host_dict);

	NUMBER_OF_APPLICATIONS = noa;

	/*allocation memoire des tables des mu et lambda *///FIXME: something dirty here!
	nb_lambda = 2 * noa * no_cpu; //pour etre large;
	nb_mu = 2 * noa * no_cpu; //Toujours pour etre large;

	lambda_table = calloc(nb_lambda, sizeof(lambda_item_t));
	for (i = 0; i < nb_lambda; i++)
		lambda_table[i] = NULL;

	mu_table = calloc(nb_mu, sizeof(mu_item_t));

	for (i = 0; i < nb_mu; i++)
		lambda_table[i] = NULL;
}

m_task_t task_create(double comp, double comm, int application,
		const char *destinataire, double *data) {
	static int number = 0;
	char buffer[128];

	sprintf(buffer, "Task (%d) sur canal %i pour %s from %s", number++,
			application, destinataire, MSG_host_self()->name);

	return MSG_task_create(buffer, comp, comm, data);
}

void task_destroy(m_task_t * task) {
	free((*task)->data);
	MSG_task_destroy(*task);
	*task = MSG_TASK_UNINITIALIZED;
}

void init_lambda(const char *identity, double data_computation_amount,
		double data_communication_amount) {
	int i = 0;
	int already_in = 0;

	while (lambda_table[i] != NULL) {

		if (strcmp(lambda_table[i]->associated_slave_name, identity) == 0) {
			already_in = 1;
			break;
		}
		i++;

	}
	if (!already_in) {

		lambda_table[i] = malloc(sizeof(struct s_lambda_item));
		lambda_item_t pointeur = lambda_table[i];

		pointeur->time = (parameters.INIT_TIME < 0) ? 0 : parameters.INIT_TIME;
		pointeur->value = parameters.LAMBDA_INIT;
		char *nom = calloc(128, sizeof(char));
		strcpy(nom, identity);
		pointeur->associated_slave_name = nom;
		pointeur->rho_sum = 0.0;
		pointeur->normalizer = 0.0;
		pointeur->normalizer_upgraded = 0.0;
		pointeur->prev_normalizer = 0.0;
		pointeur->number_of_updated_rho = 0;
		pointeur->computing_power = -1.0;
		XBT_DEBUG("My computing power : %e", pointeur->computing_power);

		pointeur->suspend_hosts_list = xbt_fifo_new();

		/*Initialisation depuis le dictionnaire si il est trouve */
		char key[128];
		char value[128];
		double quantity = 0;

		sprintf(key, "lambda %i %s", pointeur->time,
				pointeur->associated_slave_name);
		char *recuperation = NULL;
		recuperation = (char *) xbt_dict_get_or_null(readfile_dict, key);

		if (recuperation) {
			strcpy(value, recuperation);

			if (sscanf(value, "%le", &quantity) == 1) {
				pointeur->value = quantity;
			}
		}

	}
}

void ajoute_app_table(int application, int *table) {
	int i = 0;
	while (table[i] != -1) {
		i++;
	}
	table[i] = application;
}

void init_mu(const char *identity, double data_computation_amount,
		double data_communication_amount, int application_channel,
		const char *father) {
	int i = 0;
	int already_in = 0;

	while (mu_table[i] != NULL) {
		int test1 = strcmp(mu_table[i]->associated_slave_name_1, identity);
		int test2 = strcmp(mu_table[i]->associated_slave_name_2, identity);
		int test3 = strcmp(mu_table[i]->associated_slave_name_1, father);
		int test4 = strcmp(mu_table[i]->associated_slave_name_2, father);

		if ((!test1) && (!test4)) {
			already_in = 1;
			int *table_application = mu_table[i]->some_applications_at_slave_1;
			ajoute_app_table(application_channel, table_application);
			break;
		} else if ((!test2) && (!test3)) {
			already_in = 2;
			int *table_application = mu_table[i]->some_applications_at_slave_2;
			ajoute_app_table(application_channel, table_application);
			break;
		}

		i++;
	}

	if (!already_in) {

		mu_table[i] = malloc(sizeof(struct s_mu_item));
		mu_item_t pointeur = mu_table[i];
		pointeur->time = (parameters.INIT_TIME < 0) ? 0 : parameters.INIT_TIME;
		pointeur->value = parameters.MU_INIT;
		char *nom_1 = calloc(128, sizeof(char));
		strcpy(nom_1, identity);
		char *nom_2 = calloc(128, sizeof(char));
		strcpy(nom_2, father);
		pointeur->associated_slave_name_1 = nom_1;
		pointeur->associated_slave_name_2 = nom_2;
		pointeur->some_applications_at_slave_1 = calloc(NUMBER_OF_APPLICATIONS
				+ 1, sizeof(int));
		pointeur->some_applications_at_slave_2 = calloc(NUMBER_OF_APPLICATIONS
				+ 1, sizeof(int));
		int j = 0;
		for (j = 0; j < NUMBER_OF_APPLICATIONS + 1; j++) {
			pointeur->some_applications_at_slave_1[j] = -1;
			pointeur->some_applications_at_slave_2[j] = -1;
		}
		pointeur->some_applications_at_slave_1[0] = application_channel;
		pointeur->sigma_sum = 0.0;
		pointeur->normalizer = 0.0;
		pointeur->normalizer_upgraded = 0.0;
		pointeur->prev_normalizer = 0.0;
		pointeur->number_of_updated_sigma = 0;
		pointeur->bandwidth = -1.0;
		XBT_DEBUG("My bandwidth : %e", pointeur->bandwidth);
		pointeur->suspend_hosts_list = xbt_fifo_new();

		/*Initialisation depuis le dictionnaire si il est trouve */
		char key[128];
		char value[128];
		double quantity;
		sprintf(key, "mu %i %s|%s", pointeur->time,
				pointeur->associated_slave_name_1,
				pointeur->associated_slave_name_2);

		char *recuperation = NULL;
		recuperation = (char *) xbt_dict_get_or_null(readfile_dict, key);

		if (recuperation) {
			strcpy(value, recuperation);
			if (sscanf(value, "%le", &quantity) == 1) {

				pointeur->value = quantity;

			}
		}

	}
}

double fetch_lambda(const char *hostname, int temps) {
	int i;
	int found = 0;
	lambda_item_t lamb = NULL;
	/*   start_time(); */

	for (i = 0; i < nb_lambda; i++) {
		lamb = lambda_table[i];
		if (strcmp(lamb->associated_slave_name, hostname) == 0) {
			found = 1;
			break;

		}
	}

	xbt_assert(found, "Erreur dans fetch_lambda, host_name non trouve : %s",
			hostname);

	/* Insert tres propre pour synchroniser les agents */
	if (temps != lamb->time) {
		m_process_t myself = MSG_process_self();
		XBT_DEBUG("Waiting for lambda updating ... ");

		xbt_fifo_push(lamb->suspend_hosts_list, myself);

		MSG_process_suspend(myself);
		XBT_DEBUG("Stop waiting now");
	}
	xbt_assert(
			temps == lamb->time,
			"Erreur dans fetch_lambda, temps differents. Pour host : %d. Pour lambda : %d",
			temps, lamb->time);
	/*   printf("=====> %f\n", stop_time()); */
	return lamb->value;
}

/*Fonction qui verifie si un nombre est contenu dans un tableau ou pas, et qui s'arrete a number of application pour etre sur de ne pas exploser */
int is_it_in(int application, int *tableau) {
	int j = 0;
	int val = -2;

	while (val != -1) {

		val = tableau[j];
		if (application == val) {
			break;
		}

		xbt_assert(j <= NUMBER_OF_APPLICATIONS,
				"Probleme : tableau pas fini par -1");
		j++;
	}
	return val + 1;

}

double fetch_mu(const char *hostname, int application, int temps) {
	int i = 0;
	int cas = 0;
	mu_item_t mub = NULL;

	for (i = 0; i < nb_mu; i++) {
		mub = mu_table[i];
		char *name1 = (mub->associated_slave_name_1);
		char *name2 = (mub->associated_slave_name_2);
		if (strcmp(name1, hostname) == 0 && is_it_in(application,
				mub->some_applications_at_slave_1)) {
			cas = 1;
			break;
		} else if (strcmp(name2, hostname) == 0 && is_it_in(application,
				mub->some_applications_at_slave_2)) {
			cas = 2;
			break;
		}

	}

	xbt_assert(cas, "Erreur dans fetch_mu, host_name non trouve : %s",
			hostname);

	/* Insert tres propre pour synchroniser les agents */
	if (temps != mub->time) {
		m_process_t myself = MSG_process_self();
		XBT_DEBUG("Waiting for mu updating ... ");

		xbt_fifo_push(mub->suspend_hosts_list, myself);

		MSG_process_suspend(myself);
		XBT_DEBUG("Stop waiting now");
	}
	xbt_assert(
			temps == mub->time,
			"Erreur dans fetch_mu, temps differents. Pour host : %d. Pour mu : %d",
			temps, mub->time);

	return mub->value;
}

double receive_eta_from_father(m_host_t father, int application) {
	double eta;
	m_task_t task = MSG_TASK_UNINITIALIZED;
	const char channel_buf[10000];
	XBT_DEBUG("Receiving eta ...");
	XBT_DEBUG("J'ecoute : %s sur canal %i", MSG_host_get_name(father), application);
	MSG_task_receive(&task,
			CHANNEL(channel_buf,"eta", application, father, MSG_host_self()));
	MSG_task_execute(task);
	XBT_DEBUG("Receiving : %s", MSG_task_get_name(task));
	XBT_DEBUG("... eta Received");
	eta = *((double *) (task)->data);
	task_destroy(&task);
	return eta;
}

void send_eta_to_sons(m_host_t * sons, int nbsons, double eta, int application) {

	int i;
	XBT_DEBUG("Sending eta to my %i sons ...", nbsons);
	/*     start_time(); */
	for (i = 0; i < nbsons; i++) {
		m_host_t son = sons[i];
		const char *son_name = MSG_host_get_name(son);
		double *data = malloc(sizeof(double));
		const char channel_buf[10000];
		*data = eta;

		m_task_t task = task_create(AFFECTATION_QUANTITY_WORK,
				DOUBLE_QUANTITY_TO_BE_SENT, application, son_name, data);

		XBT_DEBUG("Sending : %s", MSG_task_get_name(task));
		MSG_task_send(task, CHANNEL(channel_buf,"eta", application, MSG_host_self(), son));
		XBT_DEBUG("Sending done.");
	}
	/*     printf("=====> %f\n", stop_time()); */
	XBT_DEBUG("... eta sent to all sons");
}

double receive_sigma_from_sons(m_host_t * sons, int nbsons, int application) {
	double sigma = 0.0;
	int i;
	XBT_DEBUG("Receiving sigma from  my %i sons ...", nbsons);
	for (i = 0; i < nbsons; i++) {
		m_task_t task = MSG_TASK_UNINITIALIZED;
		const char channel_buf[10000];
		XBT_DEBUG("J'ecoute : %i-%s sur canal %i", i, MSG_host_get_name(sons[i]),
				application);
		MSG_task_receive(&task, CHANNEL(channel_buf,"sigma", application, sons[i],
				MSG_host_self()));
		XBT_DEBUG("Receiving : %s", MSG_task_get_name(task));
		MSG_task_execute(task);
		sigma += *((double *) task->data);
		XBT_DEBUG("Compute of sigma done");
		task_destroy(&task);
	}
	XBT_DEBUG("... all sigma received");
	return sigma;
}

void send_sigma_to_father(m_host_t father, double sigma, int application) {

	const char *father_name = MSG_host_get_name(father);
	double *data = malloc(sizeof(double));
	const char channel_buf[10000];
	*data = sigma;
	XBT_DEBUG("Sending sigma to father ...");
	m_task_t task = task_create(AFFECTATION_QUANTITY_WORK,
			DOUBLE_QUANTITY_TO_BE_SENT, application, father_name, data);
	XBT_DEBUG("Sending : %s", MSG_task_get_name(task));
	MSG_task_send(task, CHANNEL(channel_buf,"sigma", application, MSG_host_self(), father));
	XBT_DEBUG("... sigma sent");

}

double receive_total_rho_from_father(m_host_t father, int application) {
	double rho_total;
	m_task_t task = MSG_TASK_UNINITIALIZED;
	const char channel_buf[10000];
	XBT_DEBUG("Receiving rho_total from father ...");
	XBT_DEBUG("J'ecoute : %s sur canal %i", MSG_host_get_name(father), application);
	MSG_task_receive(&task, CHANNEL(channel_buf,"rho_total", application, father,
			MSG_host_self()));
	XBT_DEBUG("Receiving : %s", MSG_task_get_name(task));
	XBT_DEBUG("... rho_total received");
	MSG_task_execute(task);
	rho_total = *((double *) task->data);
	XBT_DEBUG("Compute of rho_total done");
	task_destroy(&task);
	return rho_total;
}

void send_total_rho_to_sons(m_host_t * sons, int nbsons, double rho_total,
		int application) {

	int i;
	XBT_DEBUG("Sending rho_total to my %i sons ...", nbsons);
	for (i = 0; i < nbsons; i++) {
		m_host_t son = sons[i];
		const char *son_name = MSG_host_get_name(son);
		double *data = malloc(sizeof(double));
		const char channel_buf[10000];
		*data = rho_total;
		m_task_t task = task_create(AFFECTATION_QUANTITY_WORK,
				DOUBLE_QUANTITY_TO_BE_SENT, application, son_name, data);
		XBT_DEBUG("Sending : %s", MSG_task_get_name(task));
		MSG_task_send(task, CHANNEL(channel_buf,"rho_total", application, MSG_host_self(),
				son));
		XBT_DEBUG("Sending of rho_total done.");

	}
	XBT_DEBUG("... rho_total sent");
}

/*fonction qui donne le rho et le sigma aux tables des lambda et mu, et qui les upgrade si possible*/
void super_update_of_lambda(double rho, double computation, double rho_total,
		int application, const char *identity, const char *father_name,
		int temps) {
	/*Pour lambda */
	int i;
	int found = 0;
	lambda_item_t lamb = NULL;

	for (i = 0; i < nb_lambda; i++) {
		lamb = lambda_table[i];
		if (strcmp(lamb->associated_slave_name, identity) == 0) {
			found = 1;
			break;
		}
	}
	xbt_assert(found,
			"Erreur dans super_update_ de lambda, host_name non trouve : %s",
			identity);
	xbt_assert(
			temps == lamb->time,
			"Erreur dans super_update_ de lambda, temps differents. Pour host : %d. Pour lambda : %d",
			temps, lamb->time);

	xbt_assert(
			lamb->number_of_updated_rho < NUMBER_OF_APPLICATIONS,
			"Erreur : la taille de la file a deja atteint le NUMBER_OF_APPLICATIONS :%d",
			NUMBER_OF_APPLICATIONS);

	double newton_factor = 1.0;
	for (i = 0; i < parameters.NEWTON_HEURISTIC; i++) {
		newton_factor *= rho_total;
	}

	lamb->rho_sum += rho * computation;
	lamb->normalizer += computation * computation * newton_factor;
	if (rho > 0.0) {
		lamb->normalizer_upgraded += computation * computation * newton_factor;
	}

	lamb->number_of_updated_rho++;

	if (lamb->number_of_updated_rho == NUMBER_OF_APPLICATIONS) {

		double norm = 0.0;

		XBT_CDEBUG(alvin, "***** Update lambda [%d] *****", lamb->time);

		/*On imprime dans le dictionnaire AVANT l'update */
		//fprintf(bigfile, "lambda %d %s %g\n", lamb->time,
		//      MSG_host_self()->name, lamb->value);
		XBT_CVERB(all_vars, "lambda %d %s %g\n", lamb->time,
				MSG_host_self()->name, lamb->value);
		/*     VOIRG(rho); */
		/*     VOIRG(sigma); */
		/*     VOIRD(application); */
		/*     VOIRS(identity); */
		/*     VOIRS(father_name); */

		//lissage crade
		lamb->prev_normalizer = norm = parameters.NORM_FACTOR
				* lamb->normalizer_upgraded + (1 - parameters. NORM_FACTOR)
				* lamb->prev_normalizer;

		xbt_assert(lamb->computing_power > 0, "Oooups");

		VOIRG(lamb->rho_sum);
		VOIRG(lamb->computing_power);
		VOIRG(lamb->value);
		VOIRG(lamb->rho_sum - lamb->computing_power);

		if (lamb->rho_sum > lamb->computing_power * 1.01)
			XBT_CDEBUG(remi_bounds,
					"Contrainte CPU violee (%d):  %lf > %lf ?   ---> %lg ",
					temps, lamb->rho_sum, lamb->computing_power, lamb->rho_sum
							/ lamb->computing_power - 1.0);

		switch (parameters.SCALING_HEURISTIC) {
		case CONSTANT_STEP:
			lamb->value += parameters.GAMMA_LAMBDA * (lamb->rho_sum
					- lamb->computing_power);
			break;
		case SUM_STEP:
			lamb->value += parameters.GAMMA_LAMBDA * (lamb->rho_sum
					- lamb->computing_power) / (lamb->normalizer);
			break;
		case SMART_STEP_DIV0_MAX:
			//smart mixing of percentage and updating.
			if (parameters.GAMMA_LAMBDA * (lamb->rho_sum
					- lamb->computing_power) / (norm) > lamb->value
					* (parameters.DIV0_FACTOR - 1.0)) {
				lamb->value += parameters.GAMMA_LAMBDA * (lamb->rho_sum
						- lamb->computing_power) / (norm);
			} else {
				lamb->value = lamb->value * (parameters.DIV0_FACTOR);
			}
			break;

		default:

			if (norm > 0) {

				lamb->value += parameters.GAMMA_LAMBDA * (lamb->rho_sum
						- lamb->computing_power) / (norm);

			} else {
				switch (parameters.SCALING_HEURISTIC) {
				case SMART_STEP:
					//nothing is done
					break;
				case SMART_STEP_DIV0_PERCENTAGE:
					//dividing lambda
					lamb->value = lamb->value * parameters.DIV0_FACTOR;
					break;
				case SMART_STEP_DIV0_SUM:
					//updating with the all sum
					lamb->value += parameters.GAMMA_LAMBDA * (lamb->rho_sum
							- lamb->computing_power);
					break;
				default:
					xbt_assert(0, "Impossible happened again !");
				}

			}
		}

		// HACK by Arnaud to make sure that lambda is never negative
		if(lamb->value<0) lamb->value=0;

		//Reinit of all sums at new epoch
		lamb->number_of_updated_rho = 0;
		lamb->rho_sum = 0.0;
		lamb->normalizer = 0.0;
		lamb->normalizer_upgraded = 0.0;
		lamb->time++;
		XBT_DEBUG("Lambda updated");
		TRACE_host_variable_set_with_time((double)temps, identity, "lamb", lamb->value);

		/* Liberation des agents qui attendent */
		int ii = 0;
		int taille = xbt_fifo_size(lamb->suspend_hosts_list);
		for (ii = 0; ii < taille; ii++) {
			m_process_t myself = (m_process_t) xbt_fifo_pop(
					lamb->suspend_hosts_list);
			MSG_process_resume(myself);
		}
		XBT_DEBUG("Resuming %i processes", ii);

	}
}

static mu_item_t find_mub(int application, const char *identity) {
	/*Pour mu */
	int k = 0;
	int cas = 0;
	mu_item_t mub = NULL;
	for (k = 0; k < nb_mu; k++) {
		mub = mu_table[k];
		char *name1 = (mub->associated_slave_name_1);
		char *name2 = (mub->associated_slave_name_2);

		if (strcmp(name1, identity) == 0 && is_it_in(application,
				mub->some_applications_at_slave_1)) {
			cas = 1;
			break;
		} else if (strcmp(name2, identity) == 0 && is_it_in(application,
				mub->some_applications_at_slave_2)) {
			cas = 2;
			break;
		}

	}
	xbt_assert(cas,
			"Erreur dans super_update_ de mu , host_name non trouve : %s",
			identity);
	return mub;
}

void super_update_of_mu(double sigma, double communication, double rho_total,
		int application, const char *identity, const char *father_name,
		int temps) {
	/*Pour mu */
	int k = 0;
	mu_item_t mub = find_mub(application, identity);

	xbt_assert(
			temps == mub->time,
			"Erreur dans super_update_ de mu, temps differents. Pour host : %d. Pour mu : %d",
			temps, mub->time);

	xbt_assert(
			mub->number_of_updated_sigma < NUMBER_OF_APPLICATIONS,
			"Erreur : la taille de la file a deja atteint le NUMBER_OF_APPLICATIONS :%d",
			NUMBER_OF_APPLICATIONS);

	double newton_factor = 1.0;
	for (k = 0; k < parameters.NEWTON_HEURISTIC; k++) {
		newton_factor *= rho_total;
	}

	mub->sigma_sum += sigma * communication;
	mub->normalizer += communication * communication * newton_factor;
	if (sigma > 0.0) {
		mub->normalizer_upgraded += communication * communication
				* newton_factor;
	}

	mub->number_of_updated_sigma++;
	if (mub->number_of_updated_sigma == NUMBER_OF_APPLICATIONS) {
		XBT_CDEBUG(alvin, "**** Update mu **** ");

		double norm = 0.0;

		/*On imprime dans le dictionnaire AVANT l'update */
		//fprintf(bigfile, "mu %d %s|%s %g\n", mub->time,
		//      mub->associated_slave_name_1, mub->associated_slave_name_2,
		//      mub->value);
		XBT_CVERB(all_vars, "mu %d %s|%s %g\n", mub->time,
				mub->associated_slave_name_1, mub->associated_slave_name_2,
				mub->value);

		mub->prev_normalizer = norm = parameters.NORM_FACTOR
				* mub->normalizer_upgraded + (1 - parameters. NORM_FACTOR)
				* mub->prev_normalizer;
		xbt_assert(mub->bandwidth > 0, "Oooups");
		VOIRG(mub->sigma_sum);
		VOIRG(mub->bandwidth);
		VOIRG(mub->value);

		VOIRG(norm);
		if (mub->sigma_sum > mub->bandwidth * 1.01)
			XBT_CDEBUG(remi_bounds,
					"Contrainte BW violee (%d):  %lf > %lf ?  ---> %lg", temps,
					mub->sigma_sum, mub->bandwidth, mub->sigma_sum
							/ mub->bandwidth - 1.0);

		switch (parameters.SCALING_HEURISTIC) {
		case CONSTANT_STEP:
			mub->value += parameters.GAMMA_MU * (mub->sigma_sum
					- mub->bandwidth);
			break;
		case SUM_STEP:
			mub->value += parameters.GAMMA_MU * (mub->sigma_sum
					- mub->bandwidth) / (mub->normalizer);
			break;
		case SMART_STEP_DIV0_MAX:
			//smart mixing of percentage and updating.
			if (parameters.GAMMA_MU * (mub->sigma_sum - mub->bandwidth)
					/ (norm) > mub->value * (parameters.DIV0_FACTOR - 1.0)) {
				mub->value += parameters.GAMMA_MU * (mub->sigma_sum
						- mub->bandwidth) / (norm);
			} else {
				mub->value = mub->value * (parameters.DIV0_FACTOR);
			}
			break;

		default:

			if (norm > 0) {

				mub->value += parameters.GAMMA_MU * (mub->sigma_sum
						- mub->bandwidth) / (norm);

			} else {
				switch (parameters.SCALING_HEURISTIC) {
				case SMART_STEP:
					//nothing is done
					break;
				case SMART_STEP_DIV0_PERCENTAGE:
					//dividing mu
					mub->value = mub->value * parameters.DIV0_FACTOR;
					break;
				case SMART_STEP_DIV0_SUM:
					//updating with the all sum
					mub->value += parameters.GAMMA_MU * (mub->sigma_sum
							- mub->bandwidth);
					break;
				default:
					xbt_assert(0, "Impossible happened again !");
				}

			}
		}

		// HACK by Arnaud to make sure that mu is never negative
		if(mub->value<0) mub->value=0;



		//Reinit of all sums at new epoch
		mub->number_of_updated_sigma = 0;
		mub->sigma_sum = 0.0;
		mub->normalizer = 0.0;
		mub->normalizer_upgraded = 0.0;
		mub->time++;

		XBT_DEBUG("Mu updated");
//		TRACE_link_variable_set(mub->associated_slave_name_1,
//				mub->associated_slave_name_2, "mu", mub->value);

		trace_variable((double)temps, mub->associated_slave_name_1,
				mub->associated_slave_name_2, "mu", mub->value);

		/*Liberation des agents qui attendent */
		int ii = 0;
		int taille = xbt_fifo_size(mub->suspend_hosts_list);
		for (ii = 0; ii < taille; ii++) {
			m_process_t myself = (m_process_t) xbt_fifo_pop(
					mub->suspend_hosts_list);
			MSG_process_resume(myself);
		}
		XBT_DEBUG("Resuming %i processes", ii);
	}
}

double utility_function(double x) {
	static double MAX_VAL = 100.0;

	if (x > 0 && 1.0 / x < MAX_VAL)
		return 1.0 / x;
	else {
		MAX_VAL *= 1.1;
		return MAX_VAL;
	}
}

void wait_all_guys(void) {
  static int taille = 0;
  static smx_mutex_t mutex = NULL;
  static smx_cond_t cond = NULL;
  int num_process = xbt_dict_length(host_dict) * xbt_dict_length(appli_dict);

  if(!mutex) {
    mutex = SIMIX_req_mutex_init();
    cond =  SIMIX_req_cond_init();
  }

  SIMIX_req_mutex_lock(mutex);

  XBT_CDEBUG(alvin, "Number of waiting guys = %d", taille);

  if (taille < num_process -1) {
    taille++;
    SIMIX_req_cond_wait(cond, mutex);
    XBT_CDEBUG(alvin, "Looks like we're all here now!");
  } else if (taille == num_process-1) {
    taille = 0;
    XBT_CDEBUG(alvin, "Let's signal everyone");
    SIMIX_req_cond_broadcast(cond);
  } else {
    xbt_die("Impossible happened again !!!");
  }
  XBT_CDEBUG(alvin, "Quitting wait_all_guys");
  SIMIX_req_mutex_unlock(mutex);
}

void init_W(double W, const char *name) {
	int i = 0;
	lambda_item_t lamb;
	for (i = 0; i < nb_lambda; i++) {
		lamb = lambda_table[i];
		if (strcmp(lamb->associated_slave_name, name) == 0) {
			lamb->computing_power = W;
			break;
		}
	}
}

void init_B(double B, const char *name1, const char *name2) {
	int i = 0;
	mu_item_t mub;
	for (i = 0; i < nb_mu; i++) {
		mub = mu_table[i];

		int test1 = strcmp(mub->associated_slave_name_1, name1);
		int test2 = strcmp(mub->associated_slave_name_2, name1);
		int test3 = strcmp(mub->associated_slave_name_1, name2);
		int test4 = strcmp(mub->associated_slave_name_2, name2);

		if (((!test1) && (!test4)) || ((!test2) && (!test3))) {
			XBT_CDEBUG(alvin, "src=%s dst=%s bandwidth_value=%g", name1, name2, B);
			mub->bandwidth = B;
			break;
		}
	}
}

int grid_slave(int argc, char *argv[]) {
	const char *identity = MSG_host_get_name(MSG_host_self());
	double data_computation_amount = -1.0;
	double data_communication_amount = -1.0;
	int application_channel = -1;
	IS_MASTER test_master = -1;
	m_host_t father = NULL; /* pere inconnu */
	int nb_sons = -1;
	char color_string[3][10];
	char *color_selected;

	XBT_CDEBUG(alvin, "Parsing Arguments");
	xbt_assert(sscanf(argv[1], "%lg", &data_computation_amount) == 1,
			"Invalid Argument %s", argv[1]);
	xbt_assert(sscanf(argv[2], "%lg", &data_communication_amount) == 1,
			"Invalid Argument %s", argv[2]);
	xbt_assert(sscanf(argv[3], "%d", &application_channel) == 1,
			"Invalid Argument %s", argv[3]);
	xbt_assert(sscanf(argv[4], "%d", (int *) &test_master) == 1,
			"Invalid Argument %s", argv[4]);

	char rho_variable[100], rho_variable2[100], master_variable[100];
	snprintf(rho_variable, 100, "rho_%d", application_channel);
	snprintf(rho_variable2, 100, "RHO_%d", application_channel);
	snprintf(master_variable, 100, "master_%d", application_channel);

	{
		int pos;
		sprintf(color_string[0], "0 0 1");
		sprintf(color_string[1], "0 1 0");
		sprintf(color_string[2], "1 1 0");
		// printf("color %d : %s\n", application_channel, color_string);
		pos = application_channel % 3;
		color_selected = &color_string[ pos ][0];
	}

	if (test_master == MASTER) {
		TRACE_host_variable_declare_with_color(rho_variable, color_selected);
		TRACE_host_variable_declare_with_color(master_variable, "1 0 1");
		// time is 0 here, by definition (SaHu)
		TRACE_host_variable_set_with_time(0.0, identity, master_variable, 1);
		TRACE_link_variable_declare_with_color(rho_variable2, "1 0 0");
	}

	VOIRS(argv[1]);
	VOIRG(data_computation_amount);
	VOIRS(argv[2]);
	VOIRG(data_communication_amount);

	if (test_master == SLAVE) {
		/* pere connu dans argv[5] */
		xbt_assert(father = MSG_get_host_by_name(argv[5]),
				"Unknown father '%s'", argv[5]);
	}

	xbt_assert(sscanf(argv[6], "%d", &nb_sons) == 1, "Invalid Argument %s",
			argv[6]);

	/*tableau de m_host_contenant_les_fils contenu dans les arguments suivants grace a nb_sons */
	m_host_t *sons = calloc(nb_sons, sizeof(m_host_t));
	int i;
	for (i = 0; i < nb_sons; i++) {
		xbt_assert(sons[i] = MSG_get_host_by_name(argv[i + 7]),
				"Unknown host '%s'", argv[i + 7]);
	}

	register_application(application_channel);
	register_host();

	MSG_process_sleep(.00001);
	XBT_CDEBUG(alvin, "Creating mu and lambda tables");
	create_mu_and_lambda_table();

	/*Appel aux fonctions qui font une partie de l'initialisation des tables de lambda et de mu */

	init_lambda(identity, data_communication_amount, data_computation_amount);
	char father_name[128];
	if (test_master == SLAVE) {
		strcpy(father_name, MSG_host_get_name(father));
		init_mu(identity, data_communication_amount, data_computation_amount,
				application_channel, father_name);

	}

	XBT_CDEBUG(alvin, "Synchronize");
	/*Premiere synchronisation pour etre sur que les tables de mu et lambda ont ete crees */
	if (hosts_waiting_for_bench == NULL) {
		hosts_waiting_for_bench = xbt_fifo_new();
	}
	wait_all_guys();

	/* Ici on fait un bench pour recuperer les valeurs reelles des Wn et Bij */
	XBT_CDEBUG(alvin, "Benchmark %d", application_channel);
	double chrono = 0.0;
	double W = 0.0;
	double B = 0.0;
	double task_comp_size = 10000.0;
	double task_comm_size = 10000000000.0;

	char sprintf_buffer[64];
	m_task_t todo = NULL;

	if (application_channel == 0) {
		/* W: assez simplement : */
		sprintf(sprintf_buffer, "Bench Task W : %s", identity);
		todo = MSG_task_create(sprintf_buffer, task_comp_size, task_comm_size,
				NULL);

		chrono = MSG_get_clock();
		MSG_task_execute(todo);
		chrono = MSG_get_clock() - chrono;
		if (chrono == 0.0) {
			printf("chrono = 0 !!!! Impossible!!\n");
			exit(2);
		}
		W = task_comp_size / chrono;
		XBT_DEBUG("My CPU speed is %g", W);
		init_W(W, identity);
		MSG_task_destroy(todo);

		/*B : c'est un poil plus complique : chaque noeud (non racine envoie un message a son pere),
		 chaque pere recoit un message de ses fils. Dans le message il y a la date de l'emmission. */
		double chrono0 = 0.0;
		double chrono1 = 0.0;
		double chrono2 = 0.0;
		chrono = 0.0;

		sprintf(sprintf_buffer, "Bench Task B : %s", identity);
		/*On recoit ce qui est envoye des fils */
		for (i = 0; i < nb_sons; i++) {
			m_task_t task_r = NULL;
			double *data_r = NULL;
			const char channel_buf[10000];

			XBT_CDEBUG(alvin, "Wait for benchmack message from my son %s", sons[i]->name);
			chrono0 = MSG_get_clock();
			MSG_task_receive(&task_r, CHANNEL(channel_buf,"Benchmark", application_channel,
					sons[i], MSG_host_self()));
			XBT_CDEBUG(alvin, "Got a benchmack message from my son %s", sons[i]->name);
			chrono1 = MSG_get_clock();
			data_r = (double *) MSG_task_get_data(task_r);
			chrono2 = *data_r;
			chrono = chrono1 - ((chrono0 < chrono2) ? chrono2 : chrono0);
			xbt_assert((chrono > 0.0), "chrono = %g !!!! Impossible !!\n",
					chrono);
			B = task_comm_size / chrono;
			init_B(B, identity, sons[i]->name);

			/*Appel pour initialiser la valeur de B */
			MSG_task_destroy(task_r);
			free(data_r);
		}
		/*Si on a un pere on lui envoie un truc */
		if (test_master == SLAVE) {
			double *data_s = malloc(sizeof(double));
			m_task_t task_s = MSG_task_create(sprintf_buffer, task_comp_size,
					task_comm_size, data_s);
			const char channel_buf[10000];
			*data_s = MSG_get_clock();
			XBT_CDEBUG(alvin, "Send a benchmack message to my father %s", father->name);
			MSG_task_send(task_s, CHANNEL(channel_buf,"Benchmark", application_channel,
					MSG_host_self(), father));
			XBT_CDEBUG(alvin, "Benchmack message sent to my father %s", father->name);
		}
	}

	/*Deuxieme synchronisation : tout le monde attend l'appli 0 */
	wait_all_guys();

	XBT_CDEBUG(alvin, "Init rho");
	double rho, nonrho, price, eta, lambda, mu, rho_total, sigma;

	rho = parameters.RHO_INIT; /*initialisation de rho */
	nonrho = parameters.NONRHO_INIT; /*initialisation de nonrho */

	int real_time = (parameters.INIT_TIME < 0) ? 0 : parameters.INIT_TIME;
	/* Initialisation depuis le dictionnaire si il est trouve */
	char key[128];
	char value[128];
	double quantity;

	sprintf(key, "rho %i %s|%i", real_time, MSG_host_self()->name,
			application_channel);
	char *recuperation = NULL;

	recuperation = (char *) xbt_dict_get_or_null(readfile_dict, key);

	if (recuperation) {
		strcpy(value, (char *) xbt_dict_get_or_null(readfile_dict, key));

		if (sscanf(value, "%le", &quantity) == 1) {
			rho = quantity;

		}
	}

	sprintf(key, "rho_bis %i %s|%i", real_time, MSG_host_self()->name,
			application_channel);

	recuperation = (char *) xbt_dict_get_or_null(readfile_dict, key);

	if (recuperation) {

		strcpy(value, (char *) xbt_dict_get_or_null(readfile_dict, key));
		if (sscanf(value, "%le", &quantity) == 1) {

			nonrho = quantity;

		}
	}

	XBT_CDEBUG(alvin, "Starting Iterative algorithm");
	/* Debut de la boucle de l'algo */
	int temps;
	for (temps = real_time; temps < parameters.NB_ITERATIONS + real_time; temps++) {
		XBT_CDEBUG(alvin, "Iteration %d", temps);
		if (MSG_process_self_PID() == 1)
			VOIRD(temps);
		//fprintf(bigfile, "rho %d %s|%d %g\n", temps, MSG_host_self()->name,
		//      application_channel, rho);
		//fprintf(bigfile, "rho_bis %d %s|%d %g\n", temps, MSG_host_self()->name,
		//    application_channel, nonrho);
		XBT_CVERB(all_vars, "rho %d %s|%d %g\n", temps, MSG_host_self()->name,
				application_channel, rho);
		XBT_CVERB(all_vars, "nonrho %d %s|%d %g\n", temps, MSG_host_self()->name,
				application_channel, nonrho);
		XBT_CVERB(rho_vars, "rho %d %s|%d %g\n", temps, MSG_host_self()->name,
				application_channel, rho);

		lambda = fetch_lambda(identity, temps);

		if (test_master == SLAVE) {
			mu = fetch_mu(identity, application_channel, temps);
			eta = receive_eta_from_father(father, application_channel);
			eta = eta + mu;

		} else {
			mu = 0;
			eta = 0;
		}

		/*     if(application_channel==1 && test_master == MASTER) start_time(); */
		send_eta_to_sons(sons, nb_sons, eta, application_channel);
		/*     if(application_channel==1 && test_master == MASTER) printf("=====> %f\n", stop_time()); */

		sigma = receive_sigma_from_sons(sons, nb_sons, application_channel);
		sigma = sigma + rho;
		if (test_master == SLAVE) {
			send_sigma_to_father(father, sigma, application_channel);
			rho_total = receive_total_rho_from_father(father,
					application_channel);

		} else {
			rho_total = sigma;
		}
		send_total_rho_to_sons(sons, nb_sons, rho_total, application_channel);

		// Mise a jour des rhos
		// MSG_TASK_execute("une tache vide pour dire quand meme que on calcule");  /* Evidemment la syntaxe c'est pas ca */
		XBT_CDEBUG(alvin, "**** Update rho **** ");

		VOIRG(eta);
		VOIRG(lambda);
		price = eta * data_communication_amount + lambda
				* data_computation_amount;
		VOIRG(price);
		VOIRG(rho_total);
		VOIRG(utility_function(rho_total));
		double old_rho = rho;
		VOIRG(rho);

		int k;
		double newton_factor = (rho_total > 0.0) ? (1 / rho_total) : 0.0;
		for (k = 0; k < parameters.NEWTON_HEURISTIC; k++) {
			newton_factor *= rho_total;
		}
		rho = (1 - parameters.GAMMA_NONRHO) * rho + parameters.GAMMA_NONRHO
				* nonrho + parameters.GAMMA_RHO * (1.0 - price * rho_total)
				* newton_factor;

		VOIRG(1.0 - price * rho_total);
		VOIRG(rho - old_rho);
		rho = (rho > 0.0) ? rho : 0.0;
		VOIRG(rho);
		nonrho = (1 - parameters.GAMMA_NONRHO) * nonrho
				+ parameters.GAMMA_NONRHO * old_rho;

		//recuperation des nouveaux rho_total
		sigma = receive_sigma_from_sons(sons, nb_sons, application_channel);
		sigma = sigma + rho;
		if (test_master == SLAVE) {
			send_sigma_to_father(father, sigma, application_channel);
			rho_total = receive_total_rho_from_father(father,
					application_channel);

		} else {
			rho_total = sigma;
		}
		send_total_rho_to_sons(sons, nb_sons, rho_total, application_channel);

		//Mise a jour des lambda mu
		/*Fonction qui s'occupe de faire l'envoi du sigma et de la mise a jour si toutes les donnees sont disponibles */

		VOIRD(application_channel);
		VOIRG(rho);
		VOIRG(data_computation_amount);

		super_update_of_lambda(rho, data_computation_amount, rho_total,
				application_channel, identity, father_name, temps);

		if (test_master == SLAVE) {
			super_update_of_mu(sigma, data_communication_amount, rho_total,
					application_channel, identity, father_name, temps);
			mu_item_t mub = find_mub(application_channel, identity);
//			TRACE_link_variable_set(mub->associated_slave_name_1,
//					mub->associated_slave_name_2, rho_variable2, rho
//							* data_communication_amount);
			trace_variable((double)temps,mub->associated_slave_name_1,
					mub->associated_slave_name_2, rho_variable2, rho
							* data_communication_amount);

		}
		TRACE_host_variable_set_with_time((double)temps, identity, rho_variable, rho * data_computation_amount);
	}
	free(sons);
	return 0;
}

void trace_variable(double timestamp, const char *slave1, const char *slave2,
		const char *var_name, double value) {

	TRACE_link_srcdst_variable_set_with_time (timestamp, slave1, slave2, var_name, value);

}

