#include <msg/msg.h>
#include <xbt/fifo.h>
#include <stdio.h>

#define AFFECTATION_QUANTITY_WORK 1	/* 1 cycle processeur pour faire une affectation */
#define DOUBLE_QUANTITY_TO_BE_SENT 8	/* Quantite de donnees a envoyer pour un double */

//definition de types
typedef enum {
	REQUEST_CHANNEL = 0, GENERAL_CHANNEL, MAX_CHANNEL
} channel_t;

typedef enum {
	SLAVE = 0, MASTER = 1
} IS_MASTER;

//Define heuristic mode : which scaling for lambda and mu
typedef enum {
	CONSTANT_STEP = 0,
	SUM_STEP = 1,
	SMART_STEP = 2,
	SMART_STEP_DIV0_PERCENTAGE = 3,
	SMART_STEP_DIV0_SUM = 4,
	SMART_STEP_DIV0_MAX = 5,
} S_HEURISTIC;

typedef enum {
	NONE = 0, DIV0 = 1, NEWTON = 2
} N_HEURISTIC;

typedef struct s_request_data {
  m_host_t who_am_i;
  double computing_power;
  double bandwidth;
} *request_data_t;

typedef struct s_lambda_item {	/* type associ� a la valeur d'un lambda */
  int time;
  double value;
  char *associated_slave_name;
  double rho_sum;		/* la table des rho index� par num�ro d'application, a remplir par les agents */
  double normalizer;
  double normalizer_upgraded;
  double prev_normalizer;
  int number_of_updated_rho;	/*Le nombre d'agents qui a rempli le tableau. Quand ce nombre atteint NUMBER_OF_APPLICATIONS, on peut mettre a jour le lambda */
  double computing_power;
  xbt_fifo_t suspend_hosts_list;
  FILE *to_print;
} *lambda_item_t;

typedef struct s_mu_item {	/* type associ� a la valeur d'un mu */
  int time;
  double value;
  char *associated_slave_name_1;
  char *associated_slave_name_2;
  int *some_applications_at_slave_1;	/*Pour ces applications on considerera l'esclave 1, il faut mettre -1 a la fin */
  int *some_applications_at_slave_2;	/*Pour ces applications on considerera l'esclave 2, il faut mettre -1 a la fin */
  double sigma_sum;		/* la table des eta index� par num�ro d'application, a remplir par les agents */
  double normalizer;
  double normalizer_upgraded;
  double prev_normalizer;
  int number_of_updated_sigma;	/*Le nombre d'agents qui a rempli le tableau. Quand ce nombre atteint NUMBER_OF_APPLICATIONS, on peut mettre a jour le mu */
  double bandwidth;
  xbt_fifo_t suspend_hosts_list;
  FILE *to_print;
} *mu_item_t;



typedef struct params t_params;
struct params {
  S_HEURISTIC SCALING_HEURISTIC;
  N_HEURISTIC NEWTON_HEURISTIC;
  double GAMMA_RHO;
  double GAMMA_NONRHO;
  double GAMMA_LAMBDA;
  double GAMMA_MU;
  double RHO_INIT;
  double NONRHO_INIT;
  double LAMBDA_INIT;
  double MU_INIT;
  double NORM_FACTOR;
  double DIV0_FACTOR;
  int NB_ITERATIONS;
  int INIT_TIME;
  char OUTPUT_TYPE[255];
  int OUTPUT_START;
  char input_file[255];
  char output_file[255];
};

//Variables globales
extern double task_computation_amount;
extern double task_communication_amount;
extern double request_computation_amount;
extern double request_communication_amount;
extern double global_time;
extern t_params parameters;

// Prototypes de fonctions
void print_param(t_params param);
int main_parse(int argc, char **argv, char *platform, char *deployment);


m_task_t task_create(double comp, double comm, int application,
		     const char *destinataire, double *data);
void task_destroy(m_task_t * task);
void read_file_to_dict(int numero);
int grid_slave(int argc, char *argv[]);


void init_lambda(const char *identity, double data_computation_amount,
		 double data_communication_amount);
void ajoute_app_table(int application, int *table);
void init_mu(const char *identity, double data_computation_amount,
	     double data_communication_amount, int application_channel,
	     const char *father);
double fetch_lambda(const char *hostname, int temps);


int is_it_in(int application, int *tableau);
double fetch_mu(const char *hostname, int application, int temps);

double receive_eta_from_father(m_host_t father, int application);

void send_eta_to_sons(m_host_t * sons, int nbsons, double eta,
		      int application);

double receive_sigma_from_sons(m_host_t * sons, int nbsons,
			       int application);

void send_sigma_to_father(m_host_t father, double sigma, int application);

double receive_total_rho_from_father(m_host_t father, int application);

void send_total_rho_to_sons(m_host_t * sons, int nbsons, double rho_total,
			    int application);

void super_update_of_lambda(double rho, double normalizer,
			    double normalizer_upgraded, int application,
			    const char *identity, const char *father_name,
			    int temps);
void super_update_of_mu(double sigma, double normalizer,
			double normalizer_upgraded, int application,
			const char *identity, const char *father_name,
			int temps);

double utility_function(double x);

void print_bigfile(void);

void wait_all_guys(void);

void init_W(double W, const char *name);

void init_B(double B, const char *name1, const char *name2);
